<?php

namespace Drupal\mattermost_integration\Controller;

const TYPE_SUCCESS = 'Success';
const TYPE_ERROR = 'Error';
const TYPE_NOTICE = 'Notice';

use Drupal\comment\Entity\Comment;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Url;
use Drupal\node\Entity\Node;
use Drupal\user\Entity\User;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class EndpointController handles Outgoing Webhooks from Mattermost.
 *
 * @package Drupal\mattermost_integration\Controller
 */
class EndpointController extends ControllerBase {

  /**
   * Method to handle the incoming requests.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The value of the incoming request.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Returns a JSON string containing the result that will be returned to
   *   Mattermost.
   */
  public function requestHandler(Request $request) {
    // Decode the request body.
    $request = Json::decode($request->getContent());

    // @codingStandardsIgnoreStart
    $entities = \Drupal::entityTypeManager()->getStorage('outgoing_webhook')->loadMultiple();
    // @codingStandardsIgnoreEnd

    // Store channel IDs in a separate array.
    $channel_ids = [];
    foreach ($entities as $entity_key => $entity_value) {
      $channel_ids[$entity_key] = $entity_value->channel_id;
    }

    // Search the channel_id from the request in the stored entities.
    $search_array = array_search($request['channel_id'], $channel_ids);

    // If $search_array finds a match it will return the array key of the match.
    // If it is false it will return FALSE. The matching channel_id may be 0
    // which will also be FALSE therefore we must check for same type too.
    if ($search_array === FALSE) {
      return $this->responseBuilder(TYPE_ERROR, 'No matching channel_id was found.');
    }

    // Check if the tokens match.
    if ($entities[$search_array]->webhook_token != $request['token']) {
      return $this->responseBuilder(TYPE_ERROR, 'The webhook tokens did not match.');
    }

    // Retrieve the full post from the Mattermost server.
    $mattermost_post = $this->mattermostApiGetPost($request['post_id'], $request['team_id'], $request['channel_id']);

    // Check if the $mattermost_post contains an error or if it is empty.
    if (isset($mattermost_post['status_code'])) {
      return $this->responseBuilder(TYPE_ERROR, $mattermost_post['status_code'] . ': ' . $mattermost_post['message'] . '.');
    }
    elseif (empty($mattermost_post)) {
      return $this->responseBuilder(TYPE_ERROR, 'Received an empty object looking up the post. Is your Mattermost URL configured correctly?');
    }

    $request_post_id = $mattermost_post['order'][0];
    $request_post_value = $mattermost_post['posts'][$request_post_id];

    // Check if post contains file(s).
    $file_names = NULL;
    if (isset($request_post_value['file_ids'])) {
      foreach ($request_post_value['file_ids'] as $file_id) {
        $file_names[] = $this->mattermostApiGetFile($file_id);
      }
    }

    // Check if the request is a reply or a message by checking if the
    // root_id of the current post is empty. We get the current request ID
    // from the order array. This is an array holding the post ID of the
    // current request. We look up this post ID in the posts array and
    // check if that array has a root_id.
    // @TODO: Return a notice if files were found but couldn't be attached.
    if (empty($request_post_value['root_id'])) {
      $node = $this->createNode($request_post_value, $entities[$search_array]->content_type);

      // Check if node is created successfully.
      if ($node === FALSE) {
        return $this->responseBuilder(TYPE_ERROR, 'Could not create node.');
      }
      $url = Url::fromRoute('entity.node.canonical', ['node' => $node], ['absolute' => TRUE])->toString();

      // Attach files if any.
      if ($file_names !== NULL) {
        foreach ($file_names as $file_name) {
          $this->attachFile($node[1], $file_name);
        }
      }

      return $this->responseBuilder(TYPE_SUCCESS, $url);
    }
    else {
      $comment = $this->createReply($request_post_value, $entities[$search_array]->comment_type, $entities[$search_array]->comment_field);

      // Check if comment is created successfully.
      if ($comment === FALSE) {
        return $this->responseBuilder(TYPE_ERROR, 'Could not create comment.');
      }

      $url = Url::fromRoute('entity.comment.canonical', ['comment' => $comment], ['absolute' => TRUE])->toString();

      // Attach files if any.
      if ($file_names !== NULL) {
        foreach ($file_names as $file_name) {
          $this->attachFile($comment[1], $file_name, FALSE);
        }
      }

      return $this->responseBuilder(TYPE_SUCCESS, $url);
    }

  }

  /**
   * Function for retrieving information about a post.
   *
   * @param string $post_id
   *   The post ID of the post you want to retrieve.
   * @param string $team_id
   *   The ID of the team in which the message was posted.
   * @param string $channel_id
   *   The ID of the channel in which the message was posted.
   *
   * @return array
   *   An array with the full posts' value.
   */
  public function mattermostApiGetPost($post_id, $team_id, $channel_id) {
    // Get the configuration.
    // @codingStandardsIgnoreStart
    $config = \Drupal::config('mattermost_integration.settings')->get();
    // @codingStandardsIgnoreEnd
    $api_url = !empty($config['api_url']) ? $config['api_url'] : FALSE;
    // Get the authentication token from the state.
    // @codingStandardsIgnoreStart
    $authentication_token = \Drupal::state()->get('mattermost_authentication_token');
    // @codingStandardsIgnoreEnd
    $request_url = $api_url . '/api/v3/teams/' . $team_id . '/channels/' . $channel_id . '/posts/' . $post_id . '/get';

    $curl_handler = curl_init();
    curl_setopt($curl_handler, CURLOPT_URL, $request_url);
    curl_setopt($curl_handler, CURLOPT_RETURNTRANSFER, TRUE);
    curl_setopt($curl_handler, CURLOPT_HTTPHEADER, ['Authorization: Bearer ' . $authentication_token]);
    $response = curl_exec($curl_handler);
    curl_close($curl_handler);

    return Json::decode($response);
  }

  /**
   * Download a file to local filesystem from the Mattermost server.
   *
   * @param int $file_id
   *   The ID of the file to get.
   *
   * @return string
   *   The file name of the created file.
   */
  public function mattermostApiGetFile($file_id) {
    // Get the configuration.
    // @codingStandardsIgnoreStart
    $config = \Drupal::config('mattermost_integration.settings')->get();
    // @codingStandardsIgnoreEnd
    $api_url = !empty($config['api_url']) ? $config['api_url'] : FALSE;
    // Get the authentication token from the state.
    // @codingStandardsIgnoreStart
    $authentication_token = \Drupal::state()->get('mattermost_authentication_token');
    // @codingStandardsIgnoreEnd
    $request_url = $api_url . '/api/v3/files/' . $file_id . '/get';

    $temporary_file_directory = file_directory_temp();
    $temporary_file = $temporary_file_directory . '/mattermost_integration_file.tmp';
    $file_handle = fopen($temporary_file, 'w+');

    $curl_handler = curl_init();
    curl_setopt($curl_handler, CURLOPT_URL, $request_url);
    curl_setopt($curl_handler, CURLOPT_RETURNTRANSFER, TRUE);
    curl_setopt($curl_handler, CURLOPT_FILE, $file_handle);
    curl_setopt($curl_handler, CURLOPT_HTTPHEADER, ['Authorization: Bearer ' . $authentication_token]);
    $response = curl_exec($curl_handler);
    curl_close($curl_handler);
    fclose($file_handle);

    $metadata = Json::decode($this->mattermostApiGetFileMetadata($file_id));
    $new_file_name = $metadata['name'];
    $rename = rename($temporary_file, $temporary_file_directory . '/' . $metadata['name']);

    return $new_file_name;
  }

  /**
   * Method for getting the metadata associated to a Mattermost file.
   *
   * @param int $file_id
   *   The ID of which metadata to get.
   *
   * @return mixed
   *   The cURL response.
   */
  public function mattermostApiGetFileMetadata($file_id) {
    // Get the configuration.
    // @codingStandardsIgnoreStart
    $config = \Drupal::config('mattermost_integration.settings')->get();
    // @codingStandardsIgnoreEnd
    $api_url = !empty($config['api_url']) ? $config['api_url'] : FALSE;
    // Get the authentication token from the state.
    // @codingStandardsIgnoreStart
    $authentication_token = \Drupal::state()->get('mattermost_authentication_token');
    // @codingStandardsIgnoreEnd
    $request_url = $api_url . '/api/v3/files/' . $file_id . '/get_info';

    $curl_handler = curl_init();
    curl_setopt($curl_handler, CURLOPT_URL, $request_url);
    curl_setopt($curl_handler, CURLOPT_RETURNTRANSFER, TRUE);
    curl_setopt($curl_handler, CURLOPT_HTTPHEADER, ['Authorization: Bearer ' . $authentication_token]);
    $response = curl_exec($curl_handler);
    curl_close($curl_handler);

    return $response;

  }

  /**
   * Handler function for creating a node.
   *
   * @param array $post
   *   The full post object containing the post values.
   * @param string $content_type
   *   The type of content to create.
   *
   * @return int|bool
   *   Node ID on success or FALSE on failure.
   */
  public function createNode(array $post, $content_type) {
    $parsedown = new \Parsedown();

    $node = Node::create(['type' => $content_type]);
    $node->setTitle($this->truncateToWords($post['message'], 4, TRUE, TRUE));
    $node->setCreatedTime(substr($post['create_at'], 0, 10));
    $node->setOwnerId($this->getUserId($post['user_id']));
    $node->set('body', [
      'value' => $parsedown->text($post['message']),
      'format' => 'full_html',
    ]);
    $node->set('field_mattermost_post_id', [
      'value' => $post['id'],
    ]);
    $save_node = $node->save();
    $nid = $node->id();

    return $save_node ? $nid : FALSE;
  }

  /**
   * Method for posting a reply.
   *
   * @param array $post
   *   The post object containing the  post values.
   * @param string $comment_type
   *   The type of comment to create.
   * @param string $field_name
   *   The field name of the target entity type to attach the comment to.
   *
   * @return int|bool
   *   Comment ID on success or FALSE on failure.
   */
  public function createReply(array $post, $comment_type, $field_name) {
    $parsedown = new \Parsedown();

    $values = [
      'entity_type' => 'node',
      'entity_id' => $this->getParent($post['root_id']),
      'field_name' => $field_name,
      'comment_type' => $comment_type,
    ];

    $comment = Comment::create($values);
    $comment->setSubject($this->truncateToWords($post['message'], 2, TRUE, TRUE));
    $comment->setCreatedTime(substr($post['create_at'], 0, 10));
    $comment->setOwnerId($this->getUserId($post['user_id']));
    $comment->set('comment_body', [
      'value' => $parsedown->text($post['message']),
      'format' => 'full_html',
    ]);
    $save_comment = $comment->save();
    $cid = $comment->id();

    return $save_comment ? $cid : FALSE;
  }

  /**
   * Method for looking up a user ID from the Mattermost request.
   *
   * @param int $request_user_id
   *   The user_id found in the incoming request.
   *
   * @return int
   *   The corresponding UID, if no match found returns 0 (Anonymous).
   */
  public function getUserId($request_user_id) {
    // @codingStandardsIgnoreStart
    $user_ids = \Drupal::entityQuery('user')->execute();
    $users = User::loadMultiple($user_ids);
    // @codingStandardsIgnoreEnd

    $mattermost_user_ids = [];
    foreach ($users as $user_id => $user) {
      $mattermost_user_id = $user->get('field_mattermost_user_id')->getValue();
      if (!empty($mattermost_user_id)) {
        $mattermost_user_ids[$user_id] = $mattermost_user_id[0]['value'];
      }
    }

    // $search_array holds the User ID under which the incoming request user_id
    // is found.
    $search_array = array_search($request_user_id, $mattermost_user_ids);

    // If the user_id is not found on the Drupal site, return UID 0.
    return $search_array === FALSE ? 0 : $search_array;
  }

  /**
   * Method for getting a NID of a root_id.
   *
   * @param string $root_id
   *   The post ID which to look up.
   *
   * @return bool|mixed
   *   The NID of the matching parent node, false on no match.
   */
  public function getParent($root_id) {
    // @codingStandardsIgnoreStart
    $node_ids = \Drupal::entityQuery('node')->execute();
    $nodes = Node::loadMultiple($node_ids);
    // @codingStandardsIgnoreEnd

    $root_ids = [];
    foreach ($nodes as $node_id => $node) {
      if ($node->hasField('field_mattermost_post_id')) {
        $root_id = $node->get('field_mattermost_post_id')->getValue();
        if (!empty($root_id)) {
          $root_ids[$node_id] = $root_id;
        }
      }
    }

    $search_array = array_search($root_id, $root_ids);

    return $search_array;
  }

  /**
   * Method for truncating a string to a number of words.
   *
   * @param string $string
   *   The string to truncate.
   * @param int $word_amount
   *   The amount of words the string must be truncated to.
   * @param bool $add_ellipsis
   *   Whether or not add an ellipsis.
   * @param bool $filter_characters
   *   Whether or not to filter out characters that are not A-Z, a-z, 0-9 or
   *    hyphens.
   *
   * @return string
   *   The truncated string.
   */
  public function truncateToWords($string, $word_amount, $add_ellipsis = FALSE, $filter_characters = FALSE) {
    if ($filter_characters) {
      // Replace spaces with a placeholder so we can put the spaces back later.
      $string = str_replace(' ', '<space>', $string);

      // Strip everything but A-Z, a-z and 0-9.
      $string = preg_replace('/[^A-Za-z0-9\<space>\-]/', '', $string);

      // Restore the spaces.
      $string = str_replace('<space>', ' ', $string);
    }

    if (str_word_count($string) > $word_amount) {
      $words = explode(' ', $string);
      $string = array_slice($words, 0, $word_amount);
      $string = implode(' ', $string);
    }

    if ($add_ellipsis) {
      $string = $string . '...';
    }

    return $string;
  }

  /**
   * Method for attaching a file to a node.
   *
   * @param int $target
   *   The ID of the target node.
   * @param string $file_name
   *   The name of the file to attach.
   * @param bool $is_node
   *   Attach the file to a node, FALSE for attaching to comment.
   */
  public function attachFile($target, $file_name, $is_node = TRUE) {
    $file_data = file_get_contents(file_directory_temp() . '/' . $file_name);
    $file = file_save_data($file_data, 'public://' . $file_name);

    if ($is_node) {
      $node = Node::load($target);
      if ($node->hasField('field_mattermost_file')) {
        $files = $node->get('field_mattermost_file')->getValue();
        $files[] = $file->id();
        $node->set('field_mattermost_file', $files);
      }

      $node->save();
    }
    else {
      $comment = Comment::load($target);
      if ($comment->hasField('field_mattermost_file')) {
        $files = $comment->get('field_mattermost_file')->getValue();
        $files[] = $file->id();
        $comment->set('field_mattermost_file', $files);
      }
      $comment->save();
    }
  }

  /**
   * Method for building a JSON response.
   *
   * @param string $type
   *   The type of the message. Must be one of the TYPE_ constants.
   * @param string $message
   *   The message to respond.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Generated JsonResponse from input.
   */
  public function responseBuilder($type, $message) {
    return new JsonResponse(['text' => "$type: $message"]);
  }

}
