<?php

namespace Drupal\mattermost_integration\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\field\FieldConfigInterface;

/**
 * Form for configuring an Outgoing Webhook Entity.
 *
 * @package Drupal\mattermost_integration\Form
 */
class OutgoingWebhookForm extends EntityForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $entity = $this->entity;
    // @codingStandardsIgnoreStart
    $entity_manager = \Drupal::service('entity.manager');
    // @codingStandardsIgnoreEnd

    // Create an array with all the content types of the Drupal site.
    $content_types = [];
    foreach ($entity_manager
      ->getStorage('node_type')
      ->loadMultiple() as $content_type) {
      $content_types[$content_type->id()] = $content_type->label();
    }

    // Create an array with all the available comment types.
    $comment_types = [];
    foreach ($entity_manager
      ->getStorage('comment_type')
      ->loadMultiple() as $comment_type) {
      $comment_types[$comment_type->id()] = $comment_type->label();
    }

    $available_fields = $this->getAvailableFields($entity->content_type);
    $comment_fields = $available_fields ? $available_fields : [];

    // The label will the ID for this as well in the form of a machine_name.
    $form['id'] = [
      '#type' => 'machine_name',
      '#title' => $this->t('Label'),
      '#description' => $this->t('The label for this entity.'),
      '#maxlength' => EntityTypeInterface::BUNDLE_MAX_LENGTH,
      '#required' => TRUE,
      '#machine_name' => [
        'exists' => '\Drupal\mattermost_integration\Entity\OutgoingWebhook::load',
      ],
      '#default_value' => $entity->id(),
    ];
    $form['channel_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Channel ID'),
      '#description' => $this->t('The ID of the origin channel.'),
      '#required' => TRUE,
      '#default_value' => $entity->channel_id,
    ];
    $form['webhook_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Webhook token'),
      '#description' => $this->t('The authentication token supplied by Mattermost when a webhook is created.'),
      '#required' => TRUE,
      '#default_value' => $entity->webhook_token,
    ];
    $form['content_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Content type'),
      '#description' => $this->t('Specify which type of content this integration should create.'),
      '#required' => TRUE,
      '#options' => $content_types,
      '#default_value' => $entity->content_type,
      '#ajax' => [
        'callback' => [$this, 'updateCommentField'],
        'wrapper' => 'comment-field-wrapper',
      ],
    ];
    $form['comment_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Comment type'),
      '#description' => $this->t('Specify which type of comment this integration should create.'),
      '#required' => TRUE,
      '#options' => $comment_types,
      '#default_value' => $entity->comment_type,
    ];
    $form['wrapper'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'comment-field-wrapper'],
    ];
    $form['wrapper']['comment_field'] = [
      '#type' => 'select',
      '#title' => $this->t('Comment field'),
      '#description' => $this->t('The field on the target content type the comments should be attached to.'),
      '#required' => TRUE,
      '#options' => $comment_fields,
      '#default_value' => $entity->comment_field,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $entity = $this->entity;
    $status = $entity->save();

    if ($status) {
      drupal_set_message($this->t('Succesfully saved the entity'));
    }
    else {
      drupal_set_message($this->t('Something went wrong saving the entity, please try again later.'), 'error');
    }

    $redirect_url = Url::fromRoute('mattermost_integration.outgoing_webhooks');
    $form_state->setRedirectUrl($redirect_url);
  }

  /**
   * Method for getting the fields for a content type.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   The new form.
   */
  public function updateCommentField(array $form, FormStateInterface $form_state) {
    $content_type = $form_state->getUserInput()['content_type'];
    $fields = $this->getAvailableFields($content_type);

    $form['wrapper']['comment_field']['#options'] = $fields;

    return $form['wrapper'];
  }

  /**
   * Method for getting available fields for a content type.
   *
   * @param string $content_type
   *   The content type of which fields to get.
   *
   * @return array
   *   The available fields.
   */
  public function getAvailableFields($content_type) {
    // @codingStandardsIgnoreStart
    $entity_manager = \Drupal::service('entity.manager');
    // @codingStandardsIgnoreEnd

    $raw_fields = array_filter(
      $entity_manager->getFieldDefinitions('node', $content_type), function ($field_definition) {
        return $field_definition instanceof FieldConfigInterface;
      }
    );

    $fields = [];
    foreach ($raw_fields as $label => $raw_value) {
      $fields[$label] = $label;
    }

    return $fields;
  }

}
